import copy
import logging
from typing import (TYPE_CHECKING, Any, MutableMapping, Optional, Tuple, Union,
                    cast)

from bson import ObjectId

from nawah.config import Config

if TYPE_CHECKING:
    from nawah.classes import Attr, Query
    from nawah.types import (NawahQueryOperAnd, NawahQueryOperOr,
                             NawahQuerySpecialGroup, NawahQueryStep)

logger = logging.getLogger("nawah")


def _compile_query(
    *,
    attrs: MutableMapping[str, "Attr"],
    query: "Query",
) -> Tuple[
    Optional[int],
    Optional[int],
    MutableMapping[str, int],
    Optional[list["NawahQuerySpecialGroup"]],
    list[Any],
]:
    aggregate_prefix: list[Any] = []
    aggregate_suffix: list[Any] = []
    aggregate_query: list[Any] = [{"$match": {"$and": []}}]
    aggregate_match = aggregate_query[0]["$match"]["$and"]
    skip: Optional[int] = None
    limit: Optional[int] = None
    sort: MutableMapping[str, int] = {"_id": -1}
    group: Optional[list["NawahQuerySpecialGroup"]] = None
    logger.debug("attempting to process query: %s", query)

    query = copy.deepcopy(query)

    # Update variables per Doc Mode
    if "__deleted" not in query or query["__deleted"] is False:
        aggregate_prefix.append({"$match": {"__deleted": {"$exists": False}}})
    else:
        # This condition is expanded to allow __deleted = True, __deleted = None to have
        # del query[__deleted] be applied to both conditions
        if query["__deleted"][0] is True:
            aggregate_prefix.append({"$match": {"__deleted": {"$exists": True}}})
        del query["__deleted"][0]

    if "__create_draft" not in query or query["__create_draft"] is False:
        aggregate_prefix.append({"$match": {"__create_draft": {"$exists": False}}})
    else:
        if query["__create_draft"][0] is True:
            aggregate_prefix.append({"$match": {"__create_draft": {"$exists": True}}})
        del query["__create_draft"][0]

    if ("__update_draft" not in query and "__update_draft:$ne" not in query) or query[
        "__update_draft"
    ] is False:
        query_update_draft = False
        aggregate_prefix.append({"$match": {"__update_draft": {"$exists": False}}})
    else:
        query_update_draft = True
        aggregate_prefix.append({"$match": {"__update_draft": {"$exists": True}}})
        if "__update_draft" in query and isinstance(
            query["__update_draft"][0], ObjectId
        ):
            aggregate_prefix.append(
                {"$match": {"__update_draft": query["__update_draft"][0]}}
            )
        elif "__update_draft:$ne" in query and query["__update_draft:$ne"][0] is False:
            aggregate_prefix.append({"$match": {"__update_draft": {"$ne": False}}})
        try:
            del query["__update_draft"][0]
        except:
            del query["__update_draft:$ne"][0]

    # Update variables per Query Special Args
    if "$skip" in query:
        skip = query["$skip"]
    if "$limit" in query:
        limit = query["$limit"]
    if "$sort" in query:
        sort = query["$sort"]
    if "$group" in query:
        group = query["$group"]
    if "$search" in query:
        aggregate_prefix.insert(0, {"$match": {"$text": {"$search": query["$search"]}}})
        project_query: MutableMapping[str, Any] = {
            attr: "$" + attr for attr in attrs.keys()
        }
        project_query["_id"] = "$_id"
        project_query["__score"] = {"$meta": "textScore"}
        aggregate_suffix.append({"$project": project_query})
        aggregate_suffix.append({"$match": {"__score": {"$gt": 0.5}}})
    if "$geo_near" in query:
        aggregate_prefix.insert(
            0,
            {
                "$geoNear": {
                    "near": {
                        "type": "Point",
                        "coordinates": query["$geo_near"]["val"],
                    },
                    "distanceField": query["$geo_near"]["attr"] + ".__distance",
                    "maxDistance": query["$geo_near"]["dist"],
                    "spherical": True,
                }
            },
        )

    aggregate_match = query.pipe or []

    for step in aggregate_match:
        _compile_query_step(
            aggregate_prefix=aggregate_prefix,
            aggregate_suffix=aggregate_suffix,
            attrs=attrs,
            step=step,
        )

    if "$attrs" in query and isinstance(query["$attrs"], list):
        group_query: dict[str, Any] = {
            "_id": "$_id",
            **{
                attr: {"$first": f"${attr}"}
                for attr in query["$attrs"]
                if attr in attrs.keys()
            },
        }
        # We need to expose __update_draft value if it is queried as this refers to the original
        # doc to be updated
        if query_update_draft:
            group_query["__update_draft"] = {"$first": "$__update_draft"}
        aggregate_suffix.append({"$group": group_query})
    else:
        group_query = {
            "_id": "$_id",
            **{attr: {"$first": f"${attr}"} for attr in attrs.keys()},
        }
        # We need to expose __update_draft value if it is queried as this refers to the original
        # doc to be updated
        if query_update_draft:
            group_query["__update_draft"] = {"$first": "$__update_draft"}
        aggregate_suffix.append({"$group": group_query})

    logger.debug(
        "processed query, aggregate_prefix:%s, aggregate_suffix:%s, aggregate_match:%s",
        aggregate_prefix,
        aggregate_suffix,
        aggregate_match,
    )
    if len(aggregate_match):
        aggregate_query[0]["$match"]["$and"] = aggregate_match
    elif len(aggregate_match) == 0:
        aggregate_query = []

    aggregate_query = aggregate_prefix + aggregate_query + aggregate_suffix
    return (skip, limit, sort, group, aggregate_query)


def _compile_query_step(
    *,
    aggregate_prefix: list[Any],
    aggregate_suffix: list[Any],
    attrs: MutableMapping[str, "Attr"],
    step: Union["NawahQueryStep", "NawahQueryOperOr", "NawahQueryOperAnd"],
) -> None:
    for attr_name, attr_val in step.items():
        if isinstance(attr_val, list):
            for child_attr_val in attr_val:
                _compile_query_step(
                    aggregate_prefix=aggregate_prefix,
                    aggregate_suffix=aggregate_suffix,
                    attrs=attrs,
                    step=child_attr_val,
                )
            return

        if "." not in attr_name:
            continue

        root_attr_name = attr_name.split(".")[0]

        if root_attr_name not in attrs or not attrs[root_attr_name].extn:
            continue

        # [TODO] Check if this works with EXTN as Attr Type TYPE
        # Don't attempt to extn attr that is already extended
        lookup_query = False
        for stage in aggregate_prefix:
            if "$lookup" in stage.keys() and stage["$lookup"]["as"] == root_attr_name:
                lookup_query = True
                break
        if not lookup_query:
            extn_collection = Config.modules[
                attrs[root_attr_name].extn.module
            ].collection
            aggregate_prefix.append(
                {"$addFields": {root_attr_name: {"$toObjectId": root_attr_name}}}
            )
            aggregate_prefix.append(
                {
                    "$lookup": {
                        "from": extn_collection,
                        "localField": root_attr_name,
                        "foreignField": "_id",
                        "as": root_attr_name,
                    }
                }
            )
            aggregate_prefix.append({"$unwind": root_attr_name})
            group_query: MutableMapping[str, Any] = {
                attr: {"$first": f"${attr}"} for attr in attrs.keys()
            }
            group_query[root_attr_name] = {"$first": f"${root_attr_name}._id"}
            group_query["_id"] = "$_id"
            aggregate_suffix.append({"$group": group_query})



"""Provides 'setting' Nawah Core Module"""

from nawah.classes import Attr, Diff, Extn, Func, Module, Perm, Var

from ._callables import _create, _update

setting = Module(
    name="setting",
    desc="'setting' module module provides data type and controller for settings in Nawah "
    "eco-system. This is used by 'User' module tp provide additional user-wise settings. It "
    "also allows for global-typed settings",
    collection="setting_docs",
    attrs={
        "user": Attr.ID(desc="'_id' of 'User' doc the doc belongs to"),
        "var": Attr.STR(
            desc="Name of the setting. This is unique for every 'user' in the module"
        ),
        "val": Attr.ANY(desc="Value of the setting"),
        "val_type": Attr.ATTR(),
        "type": Attr.LITERAL(
            desc="Type of the setting. This sets whether setting is global, or belong to user, "
            "and whether use can update it or not",
            literal=["global", "user", "user_sys"],
        ),
    },
    diff=Diff(condition=lambda *_: True),
    unique_attrs=[("user", "var", "type")],
    extns={
        "val": Extn(
            module=Var.DOC("val.__extn.module"),
            attrs=Var.DOC("val.__extn.attrs"),
            force=Var.DOC("val.__extn.force"),
        ),
    },
    funcs={
        "read": Func(
            permissions=[
                Perm(privilege="admin", query_mod={"$limit": 1}),
                Perm(
                    privilege="read",
                    query_mod={
                        "user": Var.SESSION("user"),
                        # 'type': Attr.TYPE(type=attr_query_mod_type),
                        "$limit": 1,
                    },
                ),
            ],
            query_attrs=[
                {
                    "_id": Attr.ID(),
                    "type": Attr.LITERAL(literal=["global", "user", "user_sys"]),
                },
                {
                    "var": Attr.STR(),
                    "type": Attr.LITERAL(literal=["global"]),
                },
                {
                    "var": Attr.STR(),
                    "user": Attr.ID(),
                    "type": Attr.LITERAL(literal=["user", "user_sys"]),
                },
            ],
        ),
        "create": Func(
            permissions=[
                Perm(privilege="admin"),
                Perm(privilege="create", doc_mod={"type": "user"}),
            ],
            call_args={"raise_no_success": Attr.BOOL()},
            callable=_create,
        ),
        "update": Func(
            permissions=[
                Perm(privilege="admin", query_mod={"$limit": 1}),
                Perm(
                    privilege="update",
                    query_mod={
                        "type": "user",
                        "user": Var.SESSION("user"),
                        "$limit": 1,
                    },
                    doc_mod={"var": None, "val_type": None, "type": None},
                ),
            ],
            query_attrs=[
                {
                    "_id": Attr.ID(),
                    "type": Attr.LITERAL(literal=["global", "user", "user_sys"]),
                },
                {
                    "var": Attr.STR(),
                    "type": Attr.LITERAL(literal=["global"]),
                },
                {
                    "var": Attr.STR(),
                    "user": Attr.ID(),
                    "type": Attr.LITERAL(literal=["user", "user_sys"]),
                },
            ],
            call_args={"raise_no_success": Attr.BOOL()},
            callable=_update,
        ),
        "delete": Func(
            permissions=[Perm(privilege="admin", query_mod={"$limit": 1})],
            query_attrs=[{"_id": Attr.ID()}, {"var": Attr.STR()}],
        ),
        "retrieve_file": Func(
            permissions=[Perm(privilege="*", query_mod={"type": "global"})],
            get_func=True,
        ),
    },
)



import logging
from typing import TYPE_CHECKING, MutableMapping, MutableSequence, Union

from nawah.config import Config
from nawah.enums import Event

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahEnv


logger = logging.getLogger("nawah")


async def _process_file_obj(
    *,
    doc: Union["NawahDoc", MutableMapping, MutableSequence],
    env: "NawahEnv",
):
    if isinstance(doc, dict):
        doc_iter = doc.keys()
    elif isinstance(doc, list):
        doc_iter = range(len(doc))
    for j in doc_iter:
        if isinstance(doc[j], dict):
            if "__file" in doc[j].keys():
                file_id = doc[j]["__file"]
                logger.debug(
                    "Detected file in doc. Retrieving file from File module with _id: '%s'",
                    file_id,
                )
                try:
                    file_results = (
                        await Config.modules["file"]
                        .funcs["read"]
                        .callable(
                            skip_events=[Event.PERM], env=env, query=[{"_id": file_id}]
                        )
                    )
                    doc[j] = file_results["args"]["docs"][0]["file"]
                    file_results = (
                        await Config.modules["file"]
                        .funcs["delete"]
                        .callable(
                            skip_events=[Event.PERM, Event.SOFT],
                            env=env,
                            query=[{"_id": file_id}],
                        )
                    )
                    if (
                        file_results["status"] != 200
                        or file_results["args"]["count"] != 1
                    ):
                        logger.warning(
                            f"Filed to delete doc _id '{file_id}' from File module after retrieving."
                        )
                except Exception as e:
                    logger.error(f"Failed to retrieve doc _id '{file_id}', with error:")
                    logger.error(e)
                    doc[j] = None
            else:
                await _process_file_obj(doc=doc[j], env=env)
        elif type(doc[j]) == list:
            await _process_file_obj(doc=doc[j], env=env)



"""Provides 'Diff' dataclass"""

from dataclasses import dataclass
from typing import Callable


@dataclass(kw_only=True)
class Diff:
    """Diff dataclass serves role of defining Diff Instruction, which defines condition for
    creating diff doc for successful update calls. Callable for 'condition' should accept any
    number of following args only: 'mode',  'attr_name',  'attr_type',  'attr_val',  'skip_events',
    'env',  'query',  'doc',  'scope'"""

    condition: Callable



"""Provides 'session' Nawah Core Module"""

from nawah.classes import Attr, Default, Extn, Func, Module, Perm, Var

from ._callables import _auth, _reauth, _signout

session = Module(
    name="session",
    desc="'session' module provides data type and controller for sessions in Nawah eco-system. "
    "CRUD methods of the module are supposed to used for internal calls only, while methods "
    "'auth', 'reauth', and 'signout' are available for use by API as well as internal calls "
    "when needed",
    collection="session_docs",
    attrs={
        "user": Attr.ID(desc="'_id' of 'User' doc the doc belongs to"),
        "groups": Attr.LIST(
            desc="List of '_id' for every group the session is authenticated against. This attr "
            "is set by 'auth' method when called with 'groups' Doc Arg for Controller Auth "
            "Sequence",
            list=[
                Attr.ID(desc="'_id' of Group doc the session is authenticated against")
            ],
        ),
        "host_add": Attr.IP(desc="IP of the host the user used to authenticate"),
        "user_agent": Attr.STR(
            desc="User-agent of the app the user used to authenticate"
        ),
        "expiry": Attr.DATETIME(desc="Python 'datetime' ISO format of session expiry"),
        "token_hash": Attr.STR(desc="Hashed system-generated session token"),
        "create_time": Attr.DATETIME(
            desc="Python 'datetime' ISO format of the doc creation"
        ),
    },
    defaults={"groups": Default(value=[])},
    extns={"user": Extn(module="user", attrs=["*"], force=True)},
    funcs={
        "read": Func(
            permissions=[
                Perm(privilege="read", query_mod={"user": Var.SESSION("user")})
            ],
        ),
        "create": Func(permissions=[Perm(privilege="create")]),
        "update": Func(
            permissions=[
                Perm(
                    privilege="update",
                    query_mod={"user": Var.SESSION("user")},
                    doc_mod={"user": None},
                )
            ],
            query_attrs={"_id": Attr.ID()},
        ),
        "delete": Func(
            permissions=[
                Perm(privilege="__sys"),
            ],
            query_attrs={"_id": Attr.ID()},
        ),
        "auth": Func(
            permissions=[Perm(privilege="*")],
            call_args={"skip_status_check": Attr.BOOL()},
            doc_attrs=[],
            callable=_auth,
        ),
        "reauth": Func(
            permissions=[Perm(privilege="*")],
            query_attrs=[
                {
                    "_id": Attr.ID(),
                    "token": Attr.STR(),
                    "groups": Attr.LIST(list=[Attr.ID()]),
                },
                {"_id": Attr.ID(), "token": Attr.STR()},
            ],
            callable=_reauth,
        ),
        "signout": Func(
            permissions=[Perm(privilege="*")],
            query_attrs={"_id": Attr.ID()},
            callable=_signout,
        ),
    },
)



"""Provides 'create' Base Function callable"""

import asyncio
import datetime
import logging
import sys
import traceback
from functools import partial
from typing import TYPE_CHECKING, Any, MutableSequence, Optional, cast

import nawah.data as Data
from nawah.config import Config
from nawah.enums import Event
from nawah.utils import call, expand_val, validate_doc

from .exceptions import (
    DuplicateUniqueException,
    NoDocCreatedException,
    UtilityModuleDataCallException,
)

if TYPE_CHECKING:
    from asyncio.futures import Future

    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results

logger = logging.getLogger("nawah")


async def create(
    *,
    module_name: str,
    skip_events: "NawahEvents",
    env: "NawahEnv",
    doc: "NawahDoc",
    raise_no_success: Optional[bool],
) -> "Results":
    """Creates doc for a module"""

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name="create"
        )

    # Expand dot-notated keys onto dicts
    doc = expand_val(doc=doc)
    # Deleted all extra doc args
    doc = {
        attr: doc[attr]
        for attr in ["_id", *module.attrs]
        if attr in doc and doc[attr] is not None
    }
    # Append host_add, user_agent, create_time, if present in attrs
    if (
        "user" in module.attrs
        and "host_add" not in doc
        and env["session"]
        and Event.ATTRS_DOC not in skip_events
    ):
        doc["user"] = env["session"]["user"]["_id"]
    if "create_time" in module.attrs:
        doc["create_time"] = datetime.datetime.utcnow().isoformat()
    if "host_add" in module.attrs and "host_add" not in doc:
        doc["host_add"] = env["REMOTE_ADDR"]
    if "user_agent" in module.attrs and "user_agent" not in doc:
        doc["user_agent"] = env["HTTP_USER_AGENT"]
    if Event.ATTRS_DOC not in skip_events:
        # Check presence and validate all attrs in doc args
        validate_doc(
            mode="create",
            doc=doc,
            attrs=module.attrs,
        )
        # Check unique_attrs
        if module.unique_attrs:
            unique_attrs_query: MutableSequence[Any] = [[]]
            for attr in module.unique_attrs:
                if isinstance(attr, str):
                    attr = cast(str, attr)
                    unique_attrs_query[0].append({attr: doc[attr]})
                elif isinstance(attr, tuple):
                    unique_attrs_query[0].append(
                        {child_attr: doc[child_attr] for child_attr in attr}
                    )
                # [TODO] Implement use of single-item dict with LITERAL Attr Type for dynamic unique check based on doc value
            unique_attrs_query.append({"$limit": 1})
            unique_results = await call(
                "base/read",
                module_name=module_name,
                skip_events=[Event.PERM],
                env=env,
                query=unique_attrs_query,
            )
            if unique_results["args"]["count"]:
                raise DuplicateUniqueException(unique_attrs=module.unique_attrs)

    # Check for counters, to update doc
    if module.counters:
        counter_name: str
        counter_count: int
        counter_value: str
        counter_locks: dict[str, str] = {}
        counter_create: list[str] = []
        counter_update: list[str] = []

        for attr_name, counter in module.counters.items():
            # Counter.counter can be str or callable. Per type figure out lock name
            if isinstance(counter.counter, str):
                counter_name = f"__COUNTER:{counter.counter}"
            elif callable(counter.counter):
                counter_name = counter.counter(doc)

            # Obtain lock for counter_name to prevent duplicates
            counter_lock_results = await call(
                "base/obtain_lock",
                module_name="setting",
                skip_events=[Event.PERM],
                env=env,
                doc={"tags": [counter_name], "attempts": 3},
            )
            # Add lock_id to counter_locks to delete later
            counter_locks[counter_name] = counter_lock_results["args"]["docs"][0]["_id"]

            # Read counter value
            counter_results = await call(
                "setting/read",
                skip_events=[Event.PERM],
                env=env,
                query=[{"type": "global", "var": counter_name}],
            )
            # At this point, it is possible, counter was never created. Handle that scenario
            if counter_results["args"]["count"] == 0:
                # Append counter_name to counter_create so later we create it instead of updating
                # its value
                counter_create.append(counter_name)
                # And, set counter_count explicitly to 1
                counter_count = 1
            else:
                # Otherwise, add to counter_update, and add to current val
                counter_update.append(counter_name)
                counter_count = counter_results["args"]["docs"][0]["val"] + 1

            # Pass counter_count to Counter.pattern_formatter to get counter_value to be used in doc
            counter_value = counter.pattern_formatter(counter_count)
            doc[attr_name] = counter_value

    # Execute Data driver create
    results = await Data.create(env=env, collection_name=module.collection, doc=doc)

    # After successfully creating doc, create, update counters and delete locks
    # We will do this asynchronously so we don't hold response of call
    if module.counters:
        for counter_name in counter_create:
            counter_call_task = asyncio.create_task(
                call(
                    "setting/create",
                    skip_events=[Event.PERM],
                    env=env,
                    doc={
                        "type": "global",
                        "var": counter_name,
                        "val_type": {"type": "INT", "args": {}},
                        "val": 1,
                    },
                    args={"raise_no_success": True},
                )
            )
            counter_call_task.add_done_callback(
                partial(
                    _counter_call_callback,
                    env=env,
                    counter_name=counter_name,
                    counter_lock=counter_locks[counter_name],
                )
            )

        for counter_name in counter_update:
            counter_call_task = asyncio.create_task(
                call(
                    "setting/update",
                    # Skip Event ATTRS_QUERY to avoid additional requirements in query
                    skip_events=[Event.PERM, Event.ATTRS_QUERY],
                    env=env,
                    query=[{"var": counter_name}],
                    doc={
                        "val": {"$add": 1},
                    },
                    args={"raise_no_success": True},
                )
            )
            counter_call_task.add_done_callback(
                partial(
                    _counter_call_callback,
                    env=env,
                    counter_name=counter_name,
                    counter_lock=counter_locks[counter_name],
                )
            )

    # create soft action is to only return the new created doc _id.
    if Event.SOFT in skip_events:
        read_results = await call(
            "base/read",
            module_name=module_name,
            skip_events=[Event.PERM],
            env=env,
            query=[[{"_id": results["docs"][0]}]],
        )
        results = read_results["args"]

    if raise_no_success is True and results["count"] == 0:
        raise NoDocCreatedException(module_name=module_name)

    return {"status": 200, "msg": f'Created {results["count"]} docs', "args": results}


def _counter_call_callback(
    counter_call_task: "Future",
    env: "NawahEnv",
    counter_name: str,
    counter_lock: str,
):
    try:
        # Assert we have good call
        counter_call_task.result()
        # Then, delete counter_lock
        delete_lock_call_task = asyncio.create_task(
            call(
                "base/delete_lock",
                module_name="setting",
                skip_events=[Event.PERM],
                env=env,
                query=[{"_id": counter_lock}],
            )
        )
        delete_lock_call_task.add_done_callback(
            partial(
                _delete_lock_call_callback,
                counter_name=counter_name,
                counter_lock=counter_lock,
            )
        )
    except Exception:  # pylint: disable=broad-except
        exc_type, exc_value, exc_traceback = sys.exc_info()
        logger.error(
            "Call 'setting/[create|update]' for counter '%s' has failed",
            counter_name,
        )
        logger.error(
            "Lock will remain enact to prevent the counter from being reused while it is "
            "not created/updated"
        )
        logger.debug("Error traceback:")
        for line in traceback.format_exception(exc_type, exc_value, exc_traceback):
            logger.error("- %s", line)


def _delete_lock_call_callback(
    delete_lock_call_task: "Future",
    counter_name: str,
    counter_lock: str,
):
    try:
        # Assert we have good call
        delete_lock_call_task.result()
    except Exception:  # pylint: disable=broad-except
        exc_type, exc_value, exc_traceback = sys.exc_info()
        logger.error(
            "Call 'base/delete_lock' for counter '%s' with '_id' '%s' has failed. DELETE "
            "MANUALLY NOW",
            counter_name,
            counter_lock,
        )
        logger.debug("Error traceback:")
        for line in traceback.format_exception(exc_type, exc_value, exc_traceback):
            logger.error("- %s", line)



"""Provides tooling to 'setup_test' pytest market"""

from typing import TYPE_CHECKING, Any, Callable, MutableMapping, Tuple

import mock

import nawah.utils
from nawah.config import Config
from nawah.utils._config import config_module

if TYPE_CHECKING:
    from nawah.classes import L10N, Module


class MockCallRegistry:
    calls: MutableMapping[str, Callable]


class MockCall(mock.AsyncMock):
    """MockCall is used as mock for 'call' Utility. In combination with 'setup_test' it is used to
    set the return of 'call' Utility based on values returned from 'calls' callables"""

    # pylint: disable=too-many-ancestors

    def __call__(self, *args, **kwargs):
        if args[0] in MockCallRegistry.calls:

            async def _():
                return MockCallRegistry.calls[args[0]](**kwargs)

            return _()

        if args[0].split("/")[0] in Config.modules:
            return _call(args[0], **kwargs)

        raise Exception(
            f"Call to endpoint '{args[0]}' not mocked, nor its module is setup for test"
        )


_call = nawah.utils.call
nawah.utils.call = MockCall()


def setup_test(
    *,
    module: "Module" = None,
    mock_callables: bool = True,
    calls: MutableMapping[str, Callable] = None,
    l10n: MutableMapping[str, "L10N"] = None,
    vars: MutableMapping[str, Any] = None,
    types=None,
):
    """Setup Nawah runtime config for testing"""

    # pylint: disable=redefined-builtin

    Config.modules = {}
    Config.l10n = {}
    Config.vars = {}
    Config.types = {}

    if module:
        config_module(module_name=module.name, module=module)
        Config.modules[module.name] = module

        if mock_callables:
            for func in module.funcs.values():
                func.callable = mock.AsyncMock()

    if calls:
        MockCallRegistry.calls = calls

    if l10n:
        Config.l10n = l10n

    if vars:
        Config.vars = vars

    if types:
        Config.types = types



"""Provides exceptions related to 'call' Utility"""


class InvalidCallEndpointException(Exception):
    """Raised by 'call' Utility if value for arg 'endpoint' is invalid"""

    status = 400

    def __init__(self, *, endpoint):
        super().__init__(
            InvalidCallEndpointException.format_msg(endpoint=endpoint),
            {"endpoint": endpoint},
        )

    @staticmethod
    def format_msg(*, endpoint):
        """Formats exception message"""

        return f"Invliad endpoint format: '{endpoint}'"


class InvalidModuleException(Exception):
    """Raised by 'call' Utility if endpoint points to non-existent Nawah Module"""

    status = 404

    def __init__(self, *, module_name):
        super().__init__(
            InvalidModuleException.format_msg(module_name=module_name),
            {"module_name": module_name},
        )

    @staticmethod
    def format_msg(*, module_name):
        """Formats exception message"""

        return f"Nawah Module '{module_name}' is not defined"


class InvalidFuncException(Exception):
    """Raised by 'call' Utility if endpoint points to non-existent Nawah Function"""

    status = 404

    def __init__(self, *, module_name, func_name):
        super().__init__(
            InvalidFuncException.format_msg(
                module_name=module_name, func_name=func_name
            ),
            {"module_name": module_name, "func_name": func_name},
        )

    @staticmethod
    def format_msg(*, module_name, func_name):
        """Formats exception message"""

        return (
            f"Nawah Module '{module_name}' does not have Function '{func_name}' defined"
        )


class NotPermittedException(Exception):
    """Raised by 'check_permissions' Utility if failed to match Permissions Sets for current
    session user"""

    status = 403

    def __init__(self, *, module_name, func_name):
        super().__init__(
            NotPermittedException.format_msg(
                module_name=module_name, func_name=func_name
            ),
            {"module_name": module_name, "func_name": func_name},
        )

    @staticmethod
    def format_msg(*, module_name, func_name):
        """Formats exception message"""

        return f"Not permitted to access '{module_name}'.'{func_name}'"



"""Provides exceptions needed for configuring Nawah"""


class ConfigException(Exception):
    """Raised by 'config_module', 'config_app' Utility when failed to process or apply runtime
    config"""



"""Shared items required by multiple Utilities"""

from math import inf
from typing import Literal, MutableMapping, MutableSequence, TypedDict


class AttrTypeArgValidation(TypedDict, total=False):
    """Provides type-hint for 'ATTRS_TYPES_ARGS' constant"""

    type: Literal[
        "Attr", "bool", "callable", "dict", "float", "int", "list", "str", "union"
    ]
    str_pattern: str
    list_len_range: MutableSequence[float]
    list_item_type: "AttrTypeArgValidation"  # type: ignore
    dict_key_type: "AttrTypeArgValidation"  # type: ignore
    dict_val_type: "AttrTypeArgValidation"  # type: ignore
    union_types: MutableSequence["AttrTypeArgValidation"]  # type: ignore
    required: Literal[True]


# Reference to all Attr Types and bound Attr Type Args
ATTRS_TYPES_ARGS: MutableMapping[str, MutableMapping[str, AttrTypeArgValidation]] = {
    "ANY": {},
    "ATTR": {
        "types": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {"type": "str"},
        }
    },
    "ID": {},
    "STR": {"pattern": {"type": "str"}},
    "INT": {
        "ranges": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {
                "type": "list",
                "list_len_range": [2, 3],
                "list_item_type": {"type": "float"},
            },
        }
    },
    "FLOAT": {
        "ranges": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {
                "type": "list",
                "list_len_range": [2, 3],
                "list_item_type": {"type": "float"},
            },
        }
    },
    "BOOL": {},
    "LOCALE": {},
    "LOCALES": {},
    "EMAIL": {
        "allowed_domains": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {"type": "str"},
        },
        "disallowed_domains": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {"type": "str"},
        },
        "strict_matching": {"type": "bool"},
    },
    "PHONE": {
        "codes": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {"type": "str", "str_pattern": "[0-9]+"},
        }
    },
    "IP": {},
    "URI_WEB": {
        "allowed_domains": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {"type": "str"},
        },
        "disallowed_domains": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {"type": "str"},
        },
        "strict_matching": {"type": "bool"},
    },
    "DATETIME": {
        "ranges": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {
                "type": "list",
                "list_len_range": [2, 3],
                "list_item_type": {
                    "type": "str",
                    "str_pattern": r"[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?|[\-\+][0-9]+[dsmhw]",
                },
            },
        }
    },
    "DATE": {
        "ranges": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {
                "type": "list",
                "list_len_range": [2, 3],
                "list_item_type": {
                    "type": "str",
                    "str_pattern": r"[0-9]{4}-[0-9]{2}-[0-9]{2}|[\-\+][0-9]+[dsmhw]",
                },
            },
        }
    },
    "TIME": {
        "ranges": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {
                "type": "list",
                "list_len_range": [2, 3],
                "list_item_type": {
                    "type": "str",
                    "str_pattern": r"[0-9]{2}:[0-9]{2}(:[0-9]{2}(\.[0-9]{6})?)?|[\-\+][0-9]+[dsmhw]",
                },
            },
        }
    },
    "FILE": {
        "types": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {"type": "str"},
        },
        "ratio_ranges": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {
                "type": "list",
                "list_len_range": [2, 3],
                "list_item_type": {"type": "float"},
            },
        },
        "dims_ranges": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {
                "type": "list",
                "list_len_range": [2, 3],
                "list_item_type": {"type": "float"},
            },
        },
        "size_ranges": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {
                "type": "list",
                "list_len_range": [2, 3],
                "list_item_type": {"type": "float"},
            },
        },
    },
    "GEO_POINT": {},
    "LIST": {
        "list": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {"type": "Attr"},
            "required": True,
        },
        "len_range": {
            "type": "list",
            "list_len_range": [2, 3],
            "list_item_type": {"type": "float"},
        },
    },
    "KV_DICT": {
        "key": {
            "type": "Attr",
            "required": True,
        },
        "val": {
            "type": "Attr",
            "required": True,
        },
        "len_range": {
            "type": "list",
            "list_len_range": [2, 3],
            "list_item_type": {"type": "float"},
        },
        "req": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {"type": "str"},
        },
    },
    "TYPED_DICT": {
        "dict": {
            "type": "dict",
            "dict_key_type": {"type": "str"},
            "dict_val_type": {"type": "Attr"},
            "required": True,
        }
    },
    "LITERAL": {
        "literal": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {
                "type": "union",
                "union_types": [
                    {"type": "str"},
                    {"type": "int"},
                    {"type": "float"},
                    {"type": "bool"},
                ],
            },
            "required": True,
        }
    },
    "UNION": {
        "union": {
            "type": "list",
            "list_len_range": [1, inf],
            "list_item_type": {"type": "Attr"},
            "required": True,
        }
    },
    "TYPE": {
        "type": {
            "type": "union",
            "union_types": [{"type": "str"}, {"type": "callable"}],
            "required": True,
        }
    },
}



"""Provides exceptions related to validation"""


class MissingAttrException(Exception):
    """Raised by 'validate_doc' Utility if required 'attr' is missing from 'doc'"""

    status = 400

    def __init__(self, *, attr_name):
        super().__init__(
            MissingAttrException.format_msg(attr_name=attr_name),
            {"attr_name": attr_name},
        )

    @staticmethod
    def format_msg(*, attr_name):
        """Formats exception message"""

        return f"Missing attr '{attr_name}'"


class InvalidAttrException(Exception):
    """Raised by 'validate_attr' Utility if 'attr' has invalid value"""

    status = 400

    def __init__(self, *, attr_name, attr_type, val_type):
        super().__init__(
            InvalidAttrException.format_msg(
                attr_name=attr_name, attr_type=attr_type, val_type=val_type
            ),
            {"attr_name": attr_name, "attr_type": attr_type, "val_type": val_type},
        )

    @staticmethod
    def format_msg(*, attr_name, attr_type, val_type):
        """Formats exception message"""

        return f"Invalid attr '{attr_name}' of type '{val_type}' with required type '{attr_type}'"



from ._cli import cli
from ._generate import generate_models_dart, generate_models_typescript
from ._serve import serve

__all__ = ["cli", "generate_models_dart", "generate_models_typescript", "serve"]



"""Provides 'Counter' dataclass"""

from dataclasses import dataclass
from typing import Callable, Protocol, Union


@dataclass(kw_only=True)
class Counter:
    """Counter dataclass serves role as Counter Value Instruction. Callable as value for 'counter'
    should accept any number of following args only: 'doc'"""

    counter: Union[str, Callable]
    pattern_formatter: "CounterPatternFormatter"


class CounterPatternFormatter(Protocol):
    """Provides type-hint for 'pattern_formatter' callable of 'Counter'"""

    # pylint: disable=too-few-public-methods

    def __call__(
        self,
        counter_value: int,
    ) -> str:
        ...



"""Provides function to create CLI for Nawah"""

import importlib
import os
import sys

import click

from nawah import __version__

from ._create import create
from ._generate import generate_models_dart
from ._serve import serve as serve_utility


@click.group
def cli():
    """Creates CLI for Nawah"""


@cli.command
def version():
    """Print Nawah version and exit"""
    click.echo(f"Nawah v{__version__}")


@cli.command
def serve():
    """Serve Nawah app from current working directory"""
    sys.path.insert(1, os.path.abspath("."))
    module = importlib.import_module("_app")
    serve_utility(getattr(module, "app"))


@cli.group(name="generate_models")
def generate_models():
    """Generate client apps models"""


@generate_models.command
@click.option(
    "--path", prompt=" Enter path to Nawah app", type=str, default="./models/"
)
@click.option(
    "--filename",
    prompt=" Enter filename to save models to",
    type=str,
    required=False,
    default="",
)
def dart(path, filename):
    sys.path.insert(1, os.path.abspath("."))
    module = importlib.import_module("_app")
    generate_models_dart(module.app, path, filename)

    # parser_create = subparsers.add_parser('create', help='Create new Nawah app')
    # parser_create.set_defaults(func=create)
    # parser_create.add_argument(
    #     'component',
    #     help='Type of component to create',
    #     choices=['app', 'package', 'module'],
    # )

    # parser_create.add_argument(
    #     'name',
    #     type=str,
    #     help='Name of component to create',
    # )

    # parser_create.add_argument(
    #     'path',
    #     type=str,
    #     nargs='?',
    #     help='Path to create component in [default .]',
    #     default='.',
    # )

    # parser_create.add_argument(
    #     '--standalone',
    #     help='When creating Nawah package, add this flag to create standalone Python project '
    #     'for Nawah package. This allows developing Nawah package in isolation, with ability to '
    #     'publish it as Python package and use it across multiple apps. If you are intending to '
    #     'use the package in one app only, don\'t use this flag, and point path to Nawah app '
    #     'folder, to only create project sub-package',
    #     action='store_true',
    # )

    # args = parser.parse_args()

    # if args.command:
    #     args.func(args)

    # else:
    #     parser.print_help()



"""Provides 'group' Nawah Core Module"""

from nawah.classes import Attr, Default, Func, Module, Perm, Var
from nawah.config import Config

group = Module(
    name="group",
    desc="'group' module provides data type and controller for groups in Nawah eco-system",
    collection="group_docs",
    attrs={
        "user": Attr.ID(desc="'_id' of 'User' doc the doc belongs to"),
        "name": Attr.LOCALE(desc="Name of the groups as 'LOCALE'"),
        "desc": Attr.LOCALE(
            desc="Description of the group as 'LOCALE'. This can be used for dynamic generated "
            "groups that are meant to be exposed to end-users"
        ),
        "privileges": Attr.KV_DICT(
            desc="Privileges that any user is a member of the group has",
            key=Attr.STR(),
            val=Attr.LIST(list=[Attr.STR()]),
        ),
        "settings": Attr.KV_DICT(
            desc="'Setting' docs to be created, or required for members users when added to "
            "the group",
            key=Attr.STR(),
            val=Attr.ANY(),
        ),
        "create_time": Attr.DATETIME(
            desc="Python 'datetime' ISO format of the doc creation"
        ),
    },
    defaults={
        "desc": Default(value=lambda _: {locale: "" for locale in Config.locales}),
        "privileges": Default(value={}),
        "settings": Default(value={}),
    },
    funcs={
        "read": Func(
            permissions=[
                Perm(privilege="admin"),
            ],
        ),
        "create": Func(permissions=[Perm(privilege="admin")]),
        "update": Func(
            permissions=[
                Perm(privilege="admin", doc_mod={"groups": None}),
                Perm(
                    privilege="update",
                    query_mod={"_id": Var.SESSION("user")},
                    doc_mod={"groups": None, "privileges": None},
                ),
            ],
            query_attrs={"_id": Attr.ID()},
        ),
        "delete": Func(
            permissions=[
                Perm(privilege="admin"),
                Perm(privilege="delete", query_mod={"_id": Var.SESSION("user")}),
            ],
            query_attrs={"_id": Attr.ID()},
        ),
    },
)



"""Provides 'create_file' Base Function callable"""

import datetime
import logging
from typing import TYPE_CHECKING

import nawah.data as Data
from nawah.classes import Attr
from nawah.config import Config
from nawah.enums import AttrType
from nawah.utils import extract_attr, validate_doc

from ._shared import FILE_ATTRS
from .exceptions import NoDocUpdatedException, UtilityModuleDataCallException

if TYPE_CHECKING:
    from nawah.types import NawahDoc, NawahEnv, Results

logger = logging.getLogger("nawah")


async def create_file(
    *,
    module_name: str,
    env: "NawahEnv",
    doc: "NawahDoc",
) -> "Results":
    """Creates file doc for a module"""

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name="create_file"
        )
    if "_id" in doc:
        return await set_file(
            module_name=module_name,
            env=env,
            doc=doc,
            raise_no_success=True,
        )

    file_attr = extract_attr(attrs=module.attrs, path=doc["attr"])

    types_pattern = None
    if file_attr.args["types"]:
        types_pattern = "|".join(file_attr.args["types"])
        # Replace astrisks (*) with dot-astrisks to mark it as valid RegExp pattern
        types_pattern = types_pattern.replace("*", ".*")

    file_doc_attrs = {
        **FILE_ATTRS,
        "file": Attr.TYPED_DICT(
            dict={
                "name": Attr.STR(),
                "lastModified": Attr.INT(),
                "type": Attr.STR(pattern=types_pattern),
                "size": Attr.INT(),
                "content": Attr(
                    desc="__sys_attr",
                    type=AttrType.BYTES,
                    args={},
                ),
            }
        ),
    }

    file_doc = {
        "user": env["session"]["user"]["_id"],
        "doc": "000000000000000000000000",
        "attr": doc["attr"],
        "file": {
            "name": doc["name"],
            "lastModified": doc["lastModified"],
            "type": doc["type"],
            "size": len(doc["content"]),
            "content": doc["content"],
        },
        "create_time": datetime.datetime.utcnow().isoformat(),
    }

    validate_doc(mode="create", attrs=file_doc_attrs, doc=file_doc)

    # Execute Data driver create
    results = await Data.create(
        env=env, collection_name=f"{module.collection}__file", doc=file_doc
    )

    return {"status": 200, "msg": f'Created {results["count"]} files', "args": results}


async def set_file(
    *,
    module_name: str,
    env: "NawahEnv",
    doc: "NawahDoc",
    raise_no_success: bool,
) -> "Results":
    """Sets doc value of file created using 'create_file'"""
    module = Config.modules[module_name]

    update_results = await Data.update(
        env=env,
        collection_name=f"{module.collection}__file",
        docs=[doc["_id"]],
        doc={"doc": doc["doc"]},
    )

    if raise_no_success is True and update_results["count"] == 0:
        raise NoDocUpdatedException(module_name=module_name)

    return {"status": 200, "msg": "", "args": update_results}



"""Provides 'delete_file' Base Function callable"""

from typing import TYPE_CHECKING

from nawah.config import Config
from nawah.enums import AttrType, Event
from nawah.exceptions import InvalidAttrException
from nawah.utils import call

from .exceptions import (
    InvalidDeleteFileDocAttrException,
    InvalidDeleteFileIndexException,
    InvalidDeleteFileIndexValueException,
    InvalidDeleteFileMismatchNameException,
    InvalidDocException,
    UtilityModuleDataCallException,
)

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, Results


async def delete_file(
    *,
    module_name: str,
    env: "NawahEnv",
    query: "Query",
    doc: "NawahDoc",
) -> "Results":
    """Removes file from existing doc in module"""

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name="delete_file"
        )

    # [TODO] Allow use dot-notated attr path in attr query attr
    file_attr_name = query["attr"][0]
    if (
        file_attr_name not in module.attrs
        or module.attrs[file_attr_name].type != AttrType.LIST
        or not module.attrs[file_attr_name].args["list"][0].type != AttrType.FILE
    ):
        raise InvalidAttrException(
            attr_name=file_attr_name,
            attr_type=module.attrs[file_attr_name].type
            if file_attr_name in module.attrs
            else None,
            val_type=type(doc[file_attr_name]),
        )

    read_results = await call(
        "base/read",
        module_name=module_name,
        skip_events=[Event.PERM],
        env=env,
        query=[{"_id": query["_id"][0]}],
    )

    if not read_results["args"]["count"]:
        raise InvalidDocException(doc_id=query["_id"][0])

    doc = read_results["args"]["docs"][0]

    if file_attr_name not in doc or not doc[file_attr_name]:
        raise InvalidDeleteFileDocAttrException(
            doc_id=query["_id"][0], attr_name=file_attr_name
        )

    if query["index"][0] not in range(len(doc[file_attr_name])):
        raise InvalidDeleteFileIndexException(
            doc_id=query["_id"][0], attr_name=file_attr_name, index=query["index"][0]
        )

    if (
        not isinstance(doc[file_attr_name][query["index"][0]], dict)
        or "name" not in doc[file_attr_name][query["index"][0]]
    ):
        raise InvalidDeleteFileIndexValueException(
            doc_id=query["_id"][0],
            attr_name=file_attr_name,
            index=query["index"][0],
            index_val_type=type(doc[file_attr_name][query["index"][0]]),
        )

    if doc[file_attr_name][query["index"][0]]["name"] != query["name"][0]:
        raise InvalidDeleteFileMismatchNameException(
            doc_id=query["_id"][0],
            attr_name=file_attr_name,
            index=query["index"][0],
            query_file_name=query["name"][0],
            index_file_name=doc[file_attr_name][query["index"][0]]["name"],
        )

    update_results = await call(
        "base/update",
        module_name=module_name,
        skip_events=[Event.PERM],
        env=env,
        query=[{"_id": query["_id"][0]}],
        doc={file_attr_name: {"$del_val": [doc[file_attr_name][query["index"][0]]]}},
    )

    return update_results



"""Provides 'update' Base Function callable"""

import inspect
import logging
from typing import TYPE_CHECKING, Any, MutableSequence, Optional, cast

import nawah.data as Data
from nawah.config import Config
from nawah.enums import Event
from nawah.utils import call, validate_doc

from .exceptions import (
    DuplicateUniqueException,
    EmptyUpdateDocException,
    NoDocUpdatedException,
    UpdateMultiUniqueException,
    UtilityModuleDataCallException,
)

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results

logger = logging.getLogger("nawah")


async def update(
    *,
    module_name: str,
    skip_events: "NawahEvents",
    env: "NawahEnv",
    query: "Query",
    doc: "NawahDoc",
    raise_no_success: Optional[bool],
) -> "Results":
    """Updates docs for a module matching query 'query'"""

    module = Config.modules[module_name]

    if not module.collection:
        raise UtilityModuleDataCallException(
            module_name=module_name, func_name="update"
        )

    # Check presence and validate all attrs in doc args
    validate_doc(
        mode="update",
        doc=doc,
        attrs=module.attrs,
    )
    # Delete all attrs not belonging to the module
    shadow_doc = {}
    for attr_name in ["_id", *doc.keys()]:
        attr_root = attr_name.split(".")[0].split(":")[0]
        # Check root-level attr belong to module
        if attr_root not in module.attrs.keys():
            continue
        # Check attr is valid Doc Oper
        if (
            isinstance(doc[attr_name], dict)
            and doc[attr_name].keys()
            and list(doc[attr_name].keys())[0][0] == "$"
            and doc[attr_name][list(doc[attr_name].keys())[0]] is None
        ):
            continue
        # Add non-None attrs to shadow_doc
        if doc[attr_name] is not None:
            shadow_doc[attr_name] = doc[attr_name]

    doc = shadow_doc

    # Check if there is anything yet to update
    if not doc:
        if raise_no_success:
            raise EmptyUpdateDocException(module_name=module_name)

        return {"status": 200, "msg": "Nothing to update", "args": {}}
    # Find which docs are to be updated
    docs_results = await Data.read(
        env=env,
        collection_name=module.collection,
        attrs=module.attrs,
        query=query,
        skip_process=True,
    )
    # Check unique_attrs
    if module.unique_attrs:
        # If any of the unique_attrs is present in doc, and docs_results is > 1, we have duplication
        if len(docs_results["docs"]) > 1:
            unique_attrs_check = True
            for attr in module.unique_attrs:
                if isinstance(attr, str) and attr in doc:
                    unique_attrs_check = False
                    break

                if isinstance(attr, tuple):
                    for child_attr in attr:
                        if not unique_attrs_check:
                            break

                        if child_attr in doc:
                            unique_attrs_check = False
                            break

            if not unique_attrs_check:
                raise UpdateMultiUniqueException()

        # Check if any of the unique_attrs are present in doc
        if any(attr in module.unique_attrs for attr in doc):
            # Check if the doc would result in duplication after update
            unique_attrs_query: MutableSequence[Any] = [[]]
            for attr in module.unique_attrs:
                if isinstance(attr, str):
                    attr = cast(str, attr)
                    if attr in doc:
                        unique_attrs_query[0].append({attr: doc[attr]})
                elif isinstance(attr, tuple):
                    unique_attrs_query[0].append(
                        {
                            child_attr: doc[child_attr]
                            for child_attr in attr
                            if attr in doc
                        }
                    )
            unique_attrs_query.append(
                {"_id": {"$nin": [doc["_id"] for doc in docs_results["docs"]]}}
            )
            unique_attrs_query.append({"$limit": 1})
            unique_results = await call(
                "base/read",
                module_name=module_name,
                skip_events=[Event.PERM],
                env=env,
                query=unique_attrs_query,
            )
            if unique_results["args"]["count"]:
                raise DuplicateUniqueException(unique_attrs=module.unique_attrs)

    create_diff = False
    # If module has diff enabled, and Event DIFF not skipped:
    if module.diff and Event.DIFF not in skip_events:
        create_diff = True
        # Attr Type TYPE diff, call the funcion and catch InvalidAttrException
        try:
            func_params = {
                "mode": "create",
                "attr_name": "diff",
                "attr_type": module.diff,
                "attr_val": None,
                "skip_events": skip_events,
                "env": env,
                "query": query,
                "doc": doc,
                "scope": doc,
            }
            await module.diff.condition(
                **{
                    param: func_params[param]
                    for param in inspect.signature(module.diff.condition).parameters
                }
            )

        except:
            # [TODO] Implement similar error logging as in retrieve_file
            create_diff = False
            logger.debug("Skipped Diff Workflow due to failed condition")

    else:
        logger.debug(
            "Skipped Diff Workflow due to: %s, %s",
            module.diff,
            Event.DIFF not in skip_events,
        )

    results = await Data.update(
        env=env,
        collection_name=module.collection,
        docs=[doc["_id"] for doc in docs_results["docs"]],
        doc=doc,
    )

    if results["count"] and create_diff:
        for doc_result in docs_results["docs"]:
            diff_results = await Data.create(
                env=env,
                collection_name=f"{module.collection}__diff",
                doc={
                    "user": env["session"]["user"]["_id"],
                    "doc": doc_result["_id"],
                    "attrs": {
                        attr.split(".")[0]: doc_result[attr.split(".")[0]]
                        for attr in doc
                    },
                },
            )
        if not diff_results["count"]:
            logger.error("Failed to create Diff doc. Results: %s", diff_results)

    # update soft action is to only return the new created doc _id.
    if Event.SOFT in skip_events:
        read_results = await call(
            "base/read",
            module_name=module_name,
            skip_events=[Event.PERM],
            env=env,
            query=[{"_id": {"$in": [doc["_id"] for doc in docs_results["docs"]]}}],
        )
        results = read_results["args"]

    if raise_no_success is True and results["count"] == 0:
        raise NoDocUpdatedException(module_name=module_name)

    return {"status": 200, "msg": f'Updated {results["count"]} docs', "args": results}



"""Provides exceptions related to Attr Type"""


class InvalidAttrTypeException(Exception):
    """Raied by 'validate_type' Utility if Attr Type object is not valid"""

    def __init__(self, *, attr_type):
        super().__init__(
            InvalidAttrTypeException.format_msg(attr_type=attr_type),
            {"attr_type": attr_type},
        )

    @staticmethod
    def format_msg(*, attr_type):
        """Formats exception message"""

        return f"Unknown or invalid Attr Type '{attr_type}'"


class InvalidAttrTypeRefException(Exception):
    """Raied by 'validate_type' Utility if Attr Type 'TYPE' object referes to invalid type"""

    def __init__(self, *, attr_type):
        super().__init__(
            InvalidAttrTypeRefException.format_msg(attr_type=attr_type),
            {"attr_type": attr_type},
        )

    @staticmethod
    def format_msg(*, attr_type):
        """Formats exception message"""

        return f"Attr Type '{attr_type}' refers to invalid 'type'"


class InvalidAttrTypeArgException(Exception):
    """Raised by 'validate_arg' Utility if Attr Type Arg value is invalid"""

    def __init__(self, *, arg_name: str, arg_type, arg_val):
        super().__init__(
            InvalidAttrTypeArgException.format_msg(
                arg_name=arg_name, arg_type=arg_type, arg_val=arg_val
            ),
            {"arg_name": arg_name, "arg_type": arg_type, "arg_val": arg_val},
        )

    @staticmethod
    def format_msg(*, arg_name, arg_type, arg_val):
        """Formats exception message"""

        return f"Invalid Attr Type Arg for '{arg_name}' expecting type '{arg_type}' but got '{arg_val}'."


class JSONPathNotFoundException(Exception):
    """Raised by \'extract_attr\' Utility if provided didn't match any attr in 'scope'"""

    status = 400

    def __init__(self, *, scope, attr_path):
        super().__init__(
            JSONPathNotFoundException.format_msg(scope=scope, attr_path=attr_path),
            {"scope": scope, "attr_path": attr_path},
        )

    @staticmethod
    def format_msg(*, scope, attr_path):
        """Formats exception message"""

        return f"Failed to extract {attr_path} from scope: {scope}"



"""Provides 'session' Module Functions callables"""

import datetime
import secrets
from typing import TYPE_CHECKING, Optional

from passlib.hash import pbkdf2_sha512

from nawah.config import Config
from nawah.enums import Event
from nawah.exceptions import FuncException
from nawah.utils import call

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, Results


async def _auth(
    env: "NawahEnv", doc: "NawahDoc", skip_status_check: Optional[None]
) -> "Results":
    for attr in Config.modules["user"].unique_attrs:
        if attr in doc.keys():
            key = attr
            break
    user_query = [{key: doc[key], "$limit": 1}]
    if "groups" in doc and doc["groups"]:
        user_query.append(
            [{"groups": {"$in": doc["groups"]}}, {"privileges": {"*": ["*"]}}]
        )

    user_results = await call(
        "user/read",
        skip_events=[Event.PERM],
        env=env,
        query=user_query,
        args={"skip_sanitise_results": True},
    )
    if not user_results["args"]["count"] or not pbkdf2_sha512.verify(
        doc["hash"],
        user_results["args"]["docs"][0][f"{key}_hash"],
    ):
        raise FuncException(
            status=403,
            msg="Wrong auth credentials.",
            args={"code": "INVALID_CREDS"},
        )

    user = user_results["args"]["docs"][0]

    if skip_status_check is not True:
        if user["status"] in ["banned", "deleted"]:
            raise FuncException(
                status=403,
                msg=f'User is {user["status"]}.',
                args={"code": "INVALID_USER"},
            )

        if user["status"] == "disabled_password":
            raise FuncException(
                status=403,
                msg="User password is disabled.",
                args={"code": "INVALID_USER"},
            )

    token = secrets.token_urlsafe(32)
    session = {
        "user": user["_id"],
        "groups": doc["groups"] if "groups" in doc.keys() else [],
        "host_add": env["REMOTE_ADDR"],
        "user_agent": env["HTTP_USER_AGENT"],
        "expiry": (
            datetime.datetime.utcnow() + datetime.timedelta(days=30)
        ).isoformat(),
        "token_hash": pbkdf2_sha512.using(rounds=100000).hash(token),
    }

    results = await call(
        "session/create", skip_events=[Event.PERM], env=env, doc=session
    )
    if results["status"] != 200:
        return results

    session["_id"] = results["args"]["docs"][0]["_id"]
    session["user"] = user
    del session["token_hash"]
    session["token"] = token
    results["args"]["docs"][0] = session

    # read user privileges and return them
    user_results = await call(
        "user/read_privileges",
        skip_events=[Event.PERM],
        env=env,
        query=[{"_id": user["_id"]}],
    )
    if user_results["status"] != 200:
        return user_results
    results["args"]["docs"][0]["user"] = user_results["args"]["docs"][0]

    # Create CONN_AUTH Analytic doc
    # if Config.analytics_events['session_conn_auth']:
    #     analytic_doc = {
    #         'event': 'CONN_AUTH',
    #         'subevent': env['client_app'],
    #         'args': {
    #             'user': user_results['args']['docs'][0]['_id'],
    #             'session': results['args']['docs'][0]['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )
    # Create USER_AUTH Analytic doc
    # if Config.analytics_events['session_user_auth']:
    #     analytic_doc = {
    #         'event': 'USER_AUTH',
    #         'subevent': user_results['args']['docs'][0]['_id'],
    #         'args': {
    #             'session': results['args']['docs'][0]['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #             'client_app': env['client_app'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )

    return {
        "status": 200,
        "msg": "You were successfully authed.",
        "args": {"session": results["args"]["docs"][0]},
    }


async def _reauth(env: "NawahEnv", query: "Query") -> "Results":
    if str(query["_id"][0]) == "f00000000000000000000012":
        raise FuncException(
            status=400,
            msg="Reauth is not required for '__ANON' user.",
            args={"code": "ANON_REAUTH"},
        )

    session_query = [{"_id": query["_id"][0]}]
    if query["groups"][0]:
        session_query.append({"groups": {"$in": query["groups"][0]}})
    results = await call(
        "session/read", skip_events=[Event.PERM], env=env, query=session_query
    )
    if not results["args"]["count"]:
        raise FuncException(
            status=403, msg="Session is invalid.", args={"code": "INVALID_SESSION"}
        )

    if not pbkdf2_sha512.verify(
        query["token"][0], results["args"]["docs"][0]["token_hash"]
    ):
        raise FuncException(
            status=403,
            msg="Reauth token hash invalid.",
            args={"code": "INVALID_REAUTH_HASH"},
        )

    del results["args"]["docs"][0]["token_hash"]
    results["args"]["docs"][0]["token"] = query["token"][0]

    if results["args"]["docs"][0]["expiry"] < datetime.datetime.utcnow().isoformat():
        results = await call(
            "session/delete",
            skip_events=[Event.PERM, Event.SOFT],
            env=env,
            query=[{"_id": env["session"]["_id"]}],
        )
        raise FuncException(
            status=403, msg="Session had expired.", args={"code": "SESSION_EXPIRED"}
        )

    # update user's last_login timestamp
    await call(
        "user/update",
        skip_events=[Event.PERM],
        env=env,
        query=[{"_id": results["args"]["docs"][0]["user"]}],
        doc={"login_time": datetime.datetime.utcnow().isoformat()},
    )
    await call(
        "session/update",
        skip_events=[Event.PERM],
        env=env,
        query=[{"_id": results["args"]["docs"][0]["_id"]}],
        doc={
            "expiry": (
                datetime.datetime.utcnow() + datetime.timedelta(days=30)
            ).isoformat()
        },
    )
    # read user privileges and return them
    user_results = await call(
        "user/read_privileges",
        skip_events=[Event.PERM],
        env=env,
        query=[{"_id": results["args"]["docs"][0]["user"]["_id"]}],
    )
    results["args"]["docs"][0]["user"] = user_results["args"]["docs"][0]

    # Create CONN_AUTH Analytic doc
    # if Config.analytics_events['session_conn_reauth']:
    #     analytic_doc = {
    #         'event': 'CONN_REAUTH',
    #         'subevent': env['client_app'],
    #         'args': {
    #             'user': user_results['args']['docs'][0]['_id'],
    #             'session': results['args']['docs'][0]['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )
    # Create USER_AUTH Analytic doc
    # if Config.analytics_events['session_user_reauth']:
    #     analytic_doc = {
    #         'event': 'USER_REAUTH',
    #         'subevent': user_results['args']['docs'][0]['_id'],
    #         'args': {
    #             'session': results['args']['docs'][0]['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #             'client_app': env['client_app'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )

    return {
        "status": 200,
        "msg": "You were successfully reauthed.",
        "args": {"session": results["args"]["docs"][0]},
    }


async def _signout(env: "NawahEnv", query: "Query") -> "Results":
    if str(query["_id"][0]) == "f00000000000000000000012":
        raise FuncException(
            status=400,
            msg="Singout is not allowed for '__ANON' user.",
            args={"code": "ANON_SIGNOUT"},
        )

    results = await call(
        "session/read",
        skip_events=[Event.PERM],
        env=env,
        query=[{"_id": query["_id"][0]}],
    )

    if not results["args"]["count"]:
        raise FuncException(
            status=403, msg="Session is invalid.", args={"code": "INVALID_SESSION"}
        )

    results = await call(
        "session/delete",
        skip_events=[Event.PERM],
        env=env,
        query=[{"_id": env["session"]["_id"]}],
    )

    # Create CONN_AUTH Analytic doc
    # if Config.analytics_events['session_conn_deauth']:
    #     analytic_doc = {
    #         'event': 'CONN_DEAUTH',
    #         'subevent': env['client_app'],
    #         'args': {
    #             'user': env['session']['user']['_id'],
    #             'session': env['session']['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )
    # Create USER_AUTH Analytic doc
    # if Config.analytics_events['session_user_deauth']:
    #     analytic_doc = {
    #         'event': 'USER_DEAUTH',
    #         'subevent': env['session']['user']['_id'],
    #         'args': {
    #             'session': env['session']['_id'],
    #             'REMOTE_ADDR': env['REMOTE_ADDR'],
    #             'HTTP_USER_AGENT': env['HTTP_USER_AGENT'],
    #             'client_app': env['client_app'],
    #         },
    #     }
    #     analytic_results = await Registry.module('analytic').create(
    #         skip_events=[Event.PERM], env=env, doc=analytic_doc
    #     )
    #     if analytic_results['status'] != 200:
    #         logger.error(
    #             f'Failed to create \'Analytic\' doc: {analytic_doc}. Results: {analytic_results}'
    #         )

    return {
        "status": 200,
        "msg": "You are successfully signed-out.",
        "args": {"session": {"_id": "f00000000000000000000012"}},
    }



"""Python package setup for Nawah Framework"""

from distutils.command.install import install

import setuptools

from nawah import __version__


class Version(install):
    """Provides setup command to print Nawah Framework version"""

    def run(self):
        print(__version__)


class ApiLevel(install):
    """Provides setup command to print Nawah Framework API Level"""

    def run(self):
        print(".".join(__version__.split(".")[:2]))


with open("README.md", "r", encoding="UTF-8") as f:
    long_description = f.read()

with open("./requirements.txt", "r", encoding="UTF-8") as f:
    requirements = f.readlines()

with open("./dev_requirements.txt", "r", encoding="UTF-8") as f:
    dev_requirements = f.readlines()

packages = setuptools.find_packages(exclude=["tests*"])

package_data = {package: ["py.typed"] for package in packages}
package_data["nawah"].append("version.txt")

setuptools.setup(
    name="nawah",
    version=__version__,
    author="Mahmoud Abduljawad",
    author_email="mahmoud@masaar.com",
    description="Nawah framework--Rapid app development framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nawah-io/nawah_framework",
    package_data=package_data,
    packages=packages,
    project_urls={
        "Docs: Github": "https://github.com/nawah-io/nawah_docs",
        "GitHub: issues": "https://github.com/nawah-io/nawah_framework/issues",
        "GitHub: repo": "https://github.com/nawah-io/nawah_framework",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
        "Framework :: AsyncIO",
    ],
    python_requires=">=3.10.2",
    install_requires=requirements,
    extras_require={"dev": dev_requirements},
    cmdclass={
        "version": Version,
        "api_level": ApiLevel,
    },
    entry_points={
        "console_scripts": {
            "nawah = nawah.__main__:main",
        }
    },
    zip_safe=False,
)



"""Provides 'core' Nawah Package"""

from nawah.classes import Package

from .base import base
from .group import group
from .session import session
from .setting import setting
from .user import user

core = Package(
    name="core",
    api_level="2.0",
    version="2.0.0",
    modules=[base, user, group, session, setting],
)



"""Provides 'setting' Module Functions callables"""

from typing import TYPE_CHECKING

from nawah.config import Config
from nawah.enums import Event
from nawah.exceptions import FuncException
from nawah.utils import call, decode_attr_type, validate_doc

if TYPE_CHECKING:
    from nawah.classes import Query
    from nawah.types import NawahDoc, NawahEnv, NawahEvents, Results


async def _create(
    env: "NawahEnv", doc: "NawahDoc", raise_no_success: bool
) -> "Results":
    results = await call(
        "base/create",
        skip_events=[Event.PERM],
        module_name="setting",
        env=env,
        doc=doc,
        args={"raise_no_success": raise_no_success},
    )

    if results["status"] != 200:
        return results

    if doc["type"] in ["user", "user_sys"]:
        if (
            doc["user"] == env["session"]["user"]["_id"]
            and doc["var"] in Config.user_doc_settings
        ):
            env["session"]["user"][doc["var"]] = doc["val"]

    return results


async def _update(
    skip_events: "NawahEvents",
    env: "NawahEnv",
    query: "Query",
    doc: "NawahDoc",
    raise_no_success: bool,
) -> "Results":
    for attr in doc.keys():
        if attr == "val" or attr.startswith("val."):
            break
    else:
        raise FuncException(
            status=400,
            msg="Could not match doc with any of the required doc_args. Failed sets: "
            "['val': Missing]",
            args={"code": "INVALID_DOC"},
        )

    setting_results = await call(
        "setting/read",
        skip_events=[Event.PERM, *skip_events],
        env=env,
        query=query,
    )
    if not setting_results["args"]["count"]:
        raise FuncException(
            status=400, msg="Invalid Setting doc", args={"code": "INVALID_SETTING"}
        )

    setting = setting_results["args"]["docs"][0]

    if Event.ATTRS_DOC not in skip_events:
        # [DOC] Attempt to validate val against Setting val_type
        setting_val_type = decode_attr_type(encoded_attr_type=setting["val_type"])
        validate_doc(
            mode="update",
            doc=doc,
            attrs={"val": setting_val_type},
        )

    results = await call(
        "base/update",
        skip_events=[Event.PERM],
        module_name="setting",
        env=env,
        query=query,
        doc=doc,
        args={"raise_no_success": raise_no_success},
    )

    if results["status"] != 200:
        return results

    # Safely check if type in query, as Event ATTRS_QUERY might skipped confirming its existence
    if "type" in query and query["type"][0] in ["user", "user_sys"]:
        if (
            query["user"][0] == env["session"]["user"]["_id"]
            and query["var"][0] in Config.user_doc_settings
        ):
            env["session"]["user"][doc["var"]] = doc["val"]

    return results



"""Provides 'serve_app' Utility"""
import asyncio
from typing import TYPE_CHECKING

from nawah.utils._app import _run_app
from nawah.utils._config import config_app, setup_app

if TYPE_CHECKING:
    from nawah.classes import App


def serve(app_config: "App", /):
    """Takes 'App' Object and setup runtime for it to run"""

    setup_app(app_config, process_vars=True)
    asyncio.run(config_app())
    _run_app()



"""Provides 'Query' and related classes and types"""

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional, Union, cast

from bson import ObjectId

from nawah.exceptions import (InvalidQueryOperTypeException,
                              InvalidQueryStepAttrLenException,
                              InvalidQueryStepAttrTypeException,
                              InvalidQueryStepLenException,
                              InvalidQueryStepTypeException,
                              UnknownQueryOperException)

from ._jsonpath import JSONPath

if TYPE_CHECKING:
    from nawah.types import (NawahQueryOperAnd, NawahQueryOperOr,
                             NawahQuerySpecial, NawahQueryStep)

logger = logging.getLogger("nawah")

SPECIAL_ATTRS = [
    "$search",
    "$sort",
    "$skip",
    "$limit",
    "$extn",
    "$soft",
    "$attrs",
    "$group",
    "$geo_near",
]

QUERY_OPERS = {
    "$eq": (str, int, float, list, dict, ObjectId),
    "$ne": (str, int, float, list, dict, ObjectId),
    "$gt": (str, int, float),
    "$gte": (str, int, float),
    "$lt": (str, int, float),
    "$lte": (str, int, float),
    "$all": (list),
    "$in": (list),
    "$nin": (list),
    "$regex": (str),
}


@dataclass(kw_only=True)
class Query:
    pipe: Optional[
        list[Union["NawahQueryStep", "NawahQueryOperOr", "NawahQueryOperAnd"]]
    ] = None
    special: Optional["NawahQuerySpecial"] = None

    def __post_init__(self):
        if not self.pipe:
            self.pipe = []
        if not isinstance(self.pipe, list):
            raise Exception(
                f"Value for 'pipe' should be of type 'list'. Got '{type(self.pipe)}' "
                "instead"
            )

        if not self.special:
            self.special = {}
        if not isinstance(self.special, dict):
            raise Exception(
                f"Value for 'special' should be of type 'dict'. Got '{type(self.special)}' "
                "instead"
            )

        self._index = None

    @property
    def index(self) -> dict[str, list[tuple[str, Any]]]:
        if not self._index:
            self._index = {}
            _index_query_pipe(pipe=self.pipe, special=self.special, index=self._index)

        return self._index

    def reset_index(self):
        self._index = None

    def __contains__(self, o):
        if o[0] == "$":
            return o in self.special

        if ":" not in o:
            logger.warning(
                "Query.__contains__ is called with deprecated format: %s. Implicitly, appending ':$eq'",
                o,
            )
            o = f"{o}:$eq"

        return o in self.index

    def __getitem__(self, i):
        if i[0] == "$":
            return self.special[i]

        if ":" not in i:
            logger.warning(
                "Query.__getitem__ is called with deprecated format: %s. Implicitly, appending ':$eq'",
                i,
            )
            i = f"{i}:$eq"

        return QueryItem(self.index[i], self, ":*" in i)

    def __setitem__(self, i, o):
        if i[0] != "$":
            raise Exception("Query.__setitem__ used against non Query Special Attr")

        logger.warning(
            "Query.__setitem__ is called with attr: %s. This behaviour is deprecated. Use 'Query.special[special_attr]' instead",
            i,
        )
        self.special[i] = o


class QueryItem(list):
    def __init__(self, item_list, query, full_repr):
        self._item_list = item_list
        self._query = query
        super().__init__(
            [value if full_repr else value[list(value)[0]] for path, value in item_list]
        )

    def __setitem__(self, i, o):
        iterable = [i]
        if i == "*":
            iterable = range(len(self))

        for ii in iterable:
            self._item_list[ii][1][list(self._item_list[ii][1])[0]] = o

    def __delitem__(self, i):
        iterable = [i]
        if i == "*":
            iterable = range(len(self))

        for ii in iterable:
            item_path = self._item_list[ii][0].split(".")
            item_parent_path = ".".join(item_path[:-1])
            item_parent = JSONPath(item_parent_path).parse(self._query.pipe)
            del item_parent[int(item_path[-1])]

        # Reset query index, then retrieve value for index to force re-index
        self._query.reset_index()
        self._query.index  # pylint:disable=pointless-statement


def _validate_query_step(*, step):
    """Validates value of provided \'Query\' step per \'Query Arg Oper\'. If failed, raises
    \'InvalidQueryArgException\' or \'UnknownQueryArgException\' if \'Query Arg Oper\' is unknow"""

    if not isinstance(step, dict):
        raise InvalidQueryStepTypeException(step_type=type(step))

    if len(step) > 1:
        raise InvalidQueryStepLenException(step_items=tuple(step))

    step_key = list(step)[0]

    if not isinstance(step_key, str):
        # [TODO] Custom Exception
        raise Exception()

    if step_key in ("$and", "$or"):
        if not isinstance(step[step_key], list):
            # [TODO] Custom exception
            raise Exception()

        for child_step in step[step_key]:
            _validate_query_step(step=child_step)

        return

    if not isinstance(step[step_key], dict):
        raise InvalidQueryStepAttrTypeException(step_attr_type=type(step[step_key]))

    if len(step[step_key]) > 1:
        raise InvalidQueryStepAttrLenException(step_attr_items=tuple(step[step_key]))

    step_oper = list(step[step_key])[0]

    if step_oper not in QUERY_OPERS:
        raise UnknownQueryOperException(
            attr_name=step_key,
            attr_oper=step_oper,
        )

    step_val = step[step_key][step_oper]
    _validate_query_step_val(
        step=step, step_key=step_key, step_oper=step_oper, step_val=step_val
    )


def _validate_query_step_val(*, step, step_key, step_oper, step_val):
    if not isinstance(step_val, QUERY_OPERS[step_oper]):
        raise InvalidQueryOperTypeException(
            attr_name=step_key,
            attr_oper=step_oper,
            attr_type=QUERY_OPERS[step_oper],
            attr_val=step_val,
        )

    if isinstance(step_val, list):
        for i, child_step_val in enumerate(step_val):
            # User Query Operator $eq to test against all allowed values for list items
            # IMPORTANT: Appending index to step_key is to prevent nested call from checking
            # step_key again and finding the value to be _id and trying to update step_val
            _validate_query_step_val(
                step=step,
                step_key=f"{step_key}.{i}",
                step_oper="$eq",
                step_val=child_step_val,
            )

    if isinstance(step_val, dict):
        for child_step_attr, child_step_val in step_val.items():
            if not isinstance(child_step_attr, str):
                # [TODO] Custom exception
                raise Exception()

            # User Query Operator $eq to test against all allowed values for dict values
            # IMPORTANT: Appending item key to step_key is to prevent nested call from checking
            # step_key again and finding the value to be _id and trying to update step_val
            _validate_query_step_val(
                step=step,
                step_key=f"{step_key}.{child_step_attr}",
                step_oper="$eq",
                step_val=child_step_val,
            )

    if step_key == "_id":
        if isinstance(step_val, list):
            step[step_key][step_oper] = [
                ObjectId(child_step_val) for child_step_val in step_val
            ]
            return

        if not isinstance(step_val, (str, ObjectId)):
            # [TODO] Custom exception
            raise Exception()

        step[step_key][step_oper] = ObjectId(step_val)


def _index_query_pipe(
    *,
    pipe: Optional[
        list[Union["NawahQueryStep", "NawahQueryOperOr", "NawahQueryOperAnd"]]
    ],
    special: Optional["NawahQuerySpecial"],
    index=dict[str, list[tuple[str, Any]]],
    path="",
):
    if not pipe:
        pipe = []

    if not special:
        special = {}

    for i, step in enumerate(pipe):
        if len(step) > 1:
            raise Exception(f"Query step '{step}' has more than one condition")

        step_key = list(step)[0]

        if step_key in ("$or", "$and"):
            _index_query_pipe(
                pipe=cast(Any, step)[step_key],
                special=special,
                index=index,
                path=f"{path}{i}.",
            )
            continue

        step_cond = step[step_key]  # type: ignore
        step_oper = list(step_cond)[0]
        wild_attr_name = f"{step_key}:*"
        query_attr_name = f"{step_key}:{step_oper}"
        if wild_attr_name not in index:
            index[wild_attr_name] = []
        if query_attr_name not in index:
            index[query_attr_name] = []
        index[wild_attr_name].append((f"{path}{i}", step_cond))
        index[query_attr_name].append((f"{path}{i}", step_cond))



"""Provides exceptions related to 'var_value' Utility"""


class InvalidLocaleException(Exception):
    """Raied by 'var_value' Utility if 'locale' is not defined in runtime config"""

    def __init__(self, *, locale):
        super().__init__(
            InvalidLocaleException.format_msg(locale=locale),
            {"locale": locale},
        )

    @staticmethod
    def format_msg(*, locale):
        """Formats exception message"""
        return f"Invalid locale '{locale}'"


class InvalidVarException(Exception):
    """Raied by 'var_value' Utility if failed to extract value of 'var'"""

    def __init__(self, *, var):
        super().__init__(
            InvalidVarException.format_msg(var=var),
            {"var": var},
        )

    @staticmethod
    def format_msg(*, var):
        """Formats exception message"""
        return f"Invalid var '{var}'"
